// FontMetrics.cpp by katahiromz
// This file is public domain software (PDS).
#include <windows.h>
#include <windowsx.h>
#include <cstdio>

HINSTANCE g_hInstance;
TCHAR g_szSampleText[MAX_PATH] = TEXT("");

#define NUM_LINES 3
LONG g_ay[NUM_LINES];

// NOTE: TMPF_FIXED_PITCH is confusing; brain-dead
#define _TMPF_VAR_PITCH TMPF_FIXED_PITCH

typedef struct CHARSET_INFO
{
    LPCTSTR Name;
    BYTE    CharSet;
} CHARSET_INFO;

CHARSET_INFO g_charset_table[] =
{
    { TEXT("ANSI_CHARSET"), ANSI_CHARSET },
    { TEXT("DEFAULT_CHARSET"), DEFAULT_CHARSET },
    { TEXT("SYMBOL_CHARSET"), SYMBOL_CHARSET },
    { TEXT("SHIFTJIS_CHARSET"), SHIFTJIS_CHARSET },
    { TEXT("HANGEUL_CHARSET"), HANGEUL_CHARSET },
    { TEXT("HANGUL_CHARSET"), HANGUL_CHARSET },
    { TEXT("GB2312_CHARSET"), GB2312_CHARSET },
    { TEXT("CHINESEBIG5_CHARSET"), CHINESEBIG5_CHARSET },
    { TEXT("OEM_CHARSET"), OEM_CHARSET },
    { TEXT("JOHAB_CHARSET"), JOHAB_CHARSET },
    { TEXT("HEBREW_CHARSET"), HEBREW_CHARSET },
    { TEXT("ARABIC_CHARSET"), ARABIC_CHARSET },
    { TEXT("GREEK_CHARSET"), GREEK_CHARSET },
    { TEXT("TURKISH_CHARSET"), TURKISH_CHARSET },
    { TEXT("VIETNAMESE_CHARSET"), VIETNAMESE_CHARSET },
    { TEXT("THAI_CHARSET"), THAI_CHARSET },
    { TEXT("EASTEUROPE_CHARSET"), EASTEUROPE_CHARSET },
    { TEXT("RUSSIAN_CHARSET"), RUSSIAN_CHARSET },
    { TEXT("MAC_CHARSET"), MAC_CHARSET },
    { TEXT("BALTIC_CHARSET"), BALTIC_CHARSET }
};
INT g_charset_count = INT(sizeof(g_charset_table) / sizeof(g_charset_table[0]));

typedef struct QUALITY_INFO
{
    LPCTSTR Name;
    BYTE Quality;
} QUALITY_INFO;

QUALITY_INFO g_quality_table[] =
{
    { TEXT("DEFAULT_QUALITY"), DEFAULT_QUALITY },
    { TEXT("DRAFT_QUALITY"), DRAFT_QUALITY },
    { TEXT("PROOF_QUALITY"), PROOF_QUALITY },
    { TEXT("NONANTIALIASED_QUALITY"), NONANTIALIASED_QUALITY },
    { TEXT("ANTIALIASED_QUALITY"), ANTIALIASED_QUALITY },
};
INT g_quality_count = sizeof(g_quality_table) / sizeof(g_quality_table[0]);

BYTE CharSetFromName(LPCTSTR Name)
{
    for (INT i = 0; i < g_charset_count; ++i)
    {
        if (lstrcmpi(g_charset_table[i].Name, Name) == 0)
            return g_charset_table[i].CharSet;
    }
    return DEFAULT_CHARSET;
}

LPCTSTR NameFromCharSet(BYTE bCharSet)
{
    for (INT i = 0; i < g_charset_count; ++i)
    {
        if (bCharSet == g_charset_table[i].CharSet)
            return g_charset_table[i].Name;
    }
    return TEXT("");
}

LPCTSTR NameFromWeight(LONG Weight)
{
    if (Weight == 0) return TEXT("FW_DONTCARE");
    if (Weight <= FW_THIN) return TEXT("FW_THIN");
    if (Weight <= FW_EXTRALIGHT) return TEXT("FW_EXTRALIGHT");
    if (Weight <= FW_ULTRALIGHT) return TEXT("FW_ULTRALIGHT");
    if (Weight <= FW_LIGHT) return TEXT("FW_LIGHT");
    if (Weight <= FW_NORMAL) return TEXT("FW_NORMAL");
    if (Weight <= FW_REGULAR) return TEXT("FW_REGULAR");
    if (Weight <= FW_MEDIUM) return TEXT("FW_MEDIUM");
    if (Weight <= FW_SEMIBOLD) return TEXT("FW_SEMIBOLD");
    if (Weight <= FW_DEMIBOLD) return TEXT("FW_DEMIBOLD");
    if (Weight <= FW_BOLD) return TEXT("FW_BOLD");
    if (Weight <= FW_EXTRABOLD) return TEXT("FW_EXTRABOLD");
    if (Weight <= FW_ULTRABOLD) return TEXT("FW_ULTRABOLD");
    if (Weight <= FW_HEAVY) return TEXT("FW_HEAVY");
    if (Weight <= FW_BLACK) return TEXT("FW_BLACK");
    return TEXT("(unknown weight)");
}

const TCHAR *GetFontSummary(HDC hDC, LOGFONT *plf, TEXTMETRIC *ptm)
{
    static TCHAR s_sz[MAX_PATH * 3];
    TCHAR sz[64];

    wsprintf(s_sz,
        TEXT("tmHeight:%ld, tmAscent:%ld, tmDescent:%ld, tmInternalLeading:%ld, tmExternalLeading:%ld\n"),
        ptm->tmHeight, ptm->tmAscent, ptm->tmDescent,
        ptm->tmInternalLeading, ptm->tmExternalLeading);

    for (INT i = 0; i < NUM_LINES; ++i)
    {
        wsprintf(sz, TEXT("g_ay[%d]:%ld, "), i, g_ay[i]);
        lstrcat(s_sz, sz);
    }

    return s_sz;
}

extern "C"
int CALLBACK EnumFontFamExProc(
    const ENUMLOGFONTEX *   lpelf,
    const TEXTMETRIC *      lptm,
    DWORD                   FontType,
    LPARAM                  lParam)
{
    HWND hwnd = HWND(lParam);
    LPCTSTR FaceName = lpelf->elfLogFont.lfFaceName;
    if (FaceName[0] == TEXT('@'))
        return TRUE;

    INT Index = SendDlgItemMessage(hwnd, cmb1, CB_FINDSTRINGEXACT, 0, LPARAM(FaceName));
    if (Index == CB_ERR)
        SendDlgItemMessage(hwnd, cmb1, CB_ADDSTRING, 0, LPARAM(FaceName));

    return TRUE;
}

VOID UpdateFontList(HWND hwnd)
{
    LOGFONT lf;
    HDC hDC = CreateCompatibleDC(NULL);
    ZeroMemory(&lf, sizeof(lf));
    lf.lfCharSet = DEFAULT_CHARSET;
    SendDlgItemMessage(hwnd, cmb1, CB_RESETCONTENT, 0, 0);
    EnumFontFamiliesEx(hDC, &lf, FONTENUMPROC(EnumFontFamExProc), LPARAM(hwnd), 0);
    DeleteDC(hDC);
}

BOOL OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
    UpdateFontList(hwnd);

    SetDlgItemInt(hwnd, edt1, 80, FALSE);

    for (INT i = 0; i < g_charset_count; ++i)
    {
        SendDlgItemMessage(hwnd, cmb2, CB_ADDSTRING, 0, (LPARAM)g_charset_table[i].Name);
    }
    SendDlgItemMessage(hwnd, cmb2, CB_SETCURSEL, 1, 0);

    SendDlgItemMessage(hwnd, cmb4, CB_ADDSTRING, 0, (LPARAM)TEXT("px"));
    SendDlgItemMessage(hwnd, cmb4, CB_ADDSTRING, 0, (LPARAM)TEXT("pt"));
    SendDlgItemMessage(hwnd, cmb4, CB_SETCURSEL, 0, 0);

    SetDlgItemText(hwnd, edt2, g_szSampleText);

    for (INT i = 0; i < g_quality_count; ++i)
    {
        SendDlgItemMessage(hwnd, cmb6, CB_ADDSTRING, 0, (LPARAM)g_quality_table[i].Name);
    }
    SendDlgItemMessage(hwnd, cmb6, CB_SETCURSEL, 0, 0);

    return TRUE;
}

void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
    switch (id)
    {
    case cmb1:
    case cmb2:
    case cmb4:
        switch (codeNotify)
        {
        case CBN_EDITCHANGE: case CBN_SELENDOK:
            InvalidateRect(hwnd, NULL, TRUE);
            break;
        }
        break;
    case cmb6:
        switch (codeNotify)
        {
        case CBN_EDITCHANGE: case CBN_SELENDOK:
            InvalidateRect(hwnd, NULL, TRUE);
            break;
        }
        break;
    case edt1:
        if (codeNotify == EN_CHANGE)
        {
            InvalidateRect(hwnd, NULL, TRUE);
        }
        break;
    case edt2:
        if (codeNotify == EN_CHANGE)
        {
            GetDlgItemText(hwnd, edt2, g_szSampleText, MAX_PATH);
            InvalidateRect(hwnd, NULL, TRUE);
        }
        break;
    case chx1:
        InvalidateRect(hwnd, NULL, TRUE);
        break;
    case IDOK:
        EndDialog(hwnd, IDOK);
        break;
    case IDCANCEL:
        EndDialog(hwnd, IDCANCEL);
        break;
    }
}

void Line(HDC hDC, INT x0, INT y0, INT x1, INT y1)
{
    MoveToEx(hDC, x0, y0, NULL);
    LineTo(hDC, x1, y1);
}

void OnPaint(HWND hwnd)
{
    // Create GUI font
    LOGFONT lf;
    HFONT hGuiFont = (HFONT)GetStockObject(DEFAULT_GUI_FONT);
    GetObject(hGuiFont, sizeof(lf), &lf);
    lf.lfHeight = 12;
    hGuiFont = CreateFontIndirect(&lf);

    // get charset
    TCHAR szCharSet[32];
    GetDlgItemText(hwnd, cmb2, szCharSet, 32);
    BYTE bCharSet = CharSetFromName(szCharSet);

    ZeroMemory(&lf, sizeof(lf));
    lf.lfCharSet = bCharSet;

    // use pt?
    BOOL bUsePt;
    if (SendDlgItemMessage(hwnd, cmb4, CB_GETCURSEL, 0, 0) == 0)
        bUsePt = TRUE;
    else
        bUsePt = FALSE;

    // get quality
    INT iQuality = SendDlgItemMessage(hwnd, cmb6, CB_GETCURSEL, 0, 0);
    lf.lfQuality = g_quality_table[iQuality].Quality;

    PAINTSTRUCT ps;
    HDC hDC = BeginPaint(hwnd, &ps);
    if (hDC)
    {
        HGDIOBJ hFontOld;
        INT X, Y;

        SetBkMode(hDC, TRANSPARENT);

        // get height
        GetDlgItemText(hwnd, cmb1, lf.lfFaceName, LF_FACESIZE);
        INT Size = GetDlgItemInt(hwnd, edt1, NULL, TRUE);
        if (bUsePt)
            lf.lfHeight = -MulDiv(Size, GetDeviceCaps(hDC, LOGPIXELSY), 72);
        else
            lf.lfHeight = -Size;

        RECT rc;
        GetClientRect(hwnd, &rc);

        // create a font and get metrics
        TEXTMETRIC tm;
        LOGFONT lfReal;
        HFONT hFont = CreateFontIndirect(&lf);
        if (hFont)
        {
            SIZE siz;
            hFontOld = SelectObject(hDC, hFont);
            {
                TCHAR szText[MAX_PATH];
                lstrcpy(szText, g_szSampleText);

                GetTextMetrics(hDC, &tm);
                GetTextExtentPoint32(hDC, szText, lstrlen(szText), &siz);

                HBRUSH hbr = CreateSolidBrush(RGB(255, 255, 0));
                HGDIOBJ hbrOld = SelectObject(hDC, hbr);
                Rectangle(hDC, 0, 0, siz.cx, siz.cy);
                SelectObject(hDC, hbrOld);
                DeleteObject(hbr);

                X = 0;
                Y = 0;
                TextOut(hDC, X, Y, szText, lstrlen(szText));
            }
            SelectObject(hDC, hFontOld);

            GetObject(hFont, sizeof(lfReal), &lfReal);
            DeleteObject(hFont);

            g_ay[0] = tm.tmInternalLeading;
            g_ay[1] = tm.tmHeight - tm.tmDescent;
            g_ay[2] = tm.tmHeight;

            static const LPCTSTR s_asz[NUM_LINES] =
            {
                TEXT("g_ay[0] = tm.tmInternalLeading;"),
                TEXT("g_ay[1] = tm.tmHeight - tm.tmDescent;"),
                TEXT("g_ay[2] = tm.tmHeight;"),
            };

            static const COLORREF s_rgb[NUM_LINES] = 
            {
                RGB(255, 0, 0),
                RGB(0, 192, 0),
                RGB(0, 0, 255),
            };

            for (INT i = 0; i < NUM_LINES; ++i)
            {
                SetTextColor(hDC, s_rgb[i]);
                hFontOld = SelectObject(hDC, hGuiFont);
                TextOut(hDC, 0, rc.bottom - 200 + i * 18, s_asz[i], lstrlen(s_asz[i]));
                SelectObject(hDC, hFontOld);

                HPEN hPen = CreatePen(PS_SOLID, 0, s_rgb[i]);
                HGDIOBJ hPenOld = SelectObject(hDC, hPen);
                Line(hDC, rc.left, g_ay[i], rc.right, g_ay[i]);
                SelectObject(hDC, hPenOld);
                DeleteObject(hPen);
            }
        }

        // font summary
        LPCTSTR pszSummary = GetFontSummary(hDC, &lfReal, &tm);
        hFontOld = SelectObject(hDC, hGuiFont);
        rc.top = rc.bottom - 100;
        SetTextColor(hDC, RGB(0, 0, 0));
        DrawText(hDC, pszSummary, lstrlen(pszSummary), &rc, DT_WORDBREAK);
        SelectObject(hDC, hFontOld);

        EndPaint(hwnd, &ps);
    }

    DeleteObject(hGuiFont);
}

extern "C"
INT_PTR CALLBACK
DialogProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
        HANDLE_MSG(hwnd, WM_INITDIALOG, OnInitDialog);
        HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
        HANDLE_MSG(hwnd, WM_PAINT, OnPaint);
    }
    return 0;
}

extern "C"
INT WINAPI
WinMain(
    HINSTANCE   hInstance,
    HINSTANCE   hPrevInstance,
    LPSTR       lpCmdLine,
    INT         nCmdShow)
{
    g_hInstance = hInstance;
    LoadString(hInstance, 6, g_szSampleText, MAX_PATH);
    DialogBox(hInstance, MAKEINTRESOURCE(1), NULL, DialogProc);
    return 0;
}
